#include "include.h"
#include "mob_cmds.h"

/* Events */

/* This is the size of the hash table. It doesn't have to be that large - it all
 * depends on how many events you want/ how fast you want to access them
 */
#define MAX_EVENT_HASH 4096

/* #define EVENT_DEBUG */

/* mobprog stuff */
bool    prog_run (MPROG_CODE * code, CHAR_DATA * ch, const void *arg1, const void *arg2, mprog_type_t type, void *orig, const char *argument);

EVENT *event_hash[MAX_EVENT_HASH];
EVENT *global_events;

/* Table of event# -> string translation */
/* Most of the below were snipped. They serve to create a table from which
 * you can look up an event and get the name of the event
 */

#define E_STRING(e) {#e,e,TRUE },
const struct flag_type event_types[] = {
    E_STRING(event_misc)
        E_STRING(event_temp)
        E_STRING(event_mprog)

        /* ... */
    {""}
};



/* Remove event 'e' from event list event_list
 * This will remove the event from a local (e.g. on an object, mob, room)
 * linked list (threaded via next_local)
 */
void remove_event_local(EVENT **event_list, EVENT *e)
{
	EVENT *e_prev;
	
	if (*event_list == e)
		*event_list = e->next_local;
	else
	{
		for (e_prev = *event_list;e_prev && e_prev->next_local != e; e_prev = e_prev->next_local)
			;
		
		e_prev->next_local = e->next_local;
	}
	
	if (e->type == event_temp && e->data.temp.remove)
		*(e->data.temp.remove) = NULL;
}

/* Some event funtions for misc (persistant) event */
EventCallbackFun decay_object_event, healing_event;

/* Misc event callback functions (if any)
 * They have to have a table, since they can save in pfiles (and we
 * can't put the function pointer there)
 *
 */
EventCallbackFun *misc_event_callbacks[max_misc_event]
    =
{
	[emisc_object_decay]			decay_object_event,
	[emisc_heal]					healing_event,
    [emisc_burning_object]			burning_object_event,
    [emisc_narcolepsy]              narcolepsy_event,
    [emisc_skill_gain]              skill_gain_event,
    [emisc_skill_decay]             skill_decay_event,
    [emisc_pk_explanation]          pk_explanation_event
}
;

/* figure out what this ptr really points at
 * This uses magic numbers to figure out if this object is a mobile, character or what
 *
 */
const char *ptr_string(void *ptr)
{
    char *buf;
	
	if (!ptr)
		return "<null>";

    buf=BUF(256); /* allocate static buffer space */
	switch(*(unsigned char*)ptr)
	{
		case MAGIC_CHAR:
		{
			CHAR_DATA *ch = ptr;
			if (IS_NPC(ch))
				snprintf(buf, 256, "[%5d] %s", ch->pIndexData->vnum, nocolor(ch->short_descr));
			else
				snprintf(buf, 256, "[ PLR ] %s", ch->name);
			break;
		}
		
		case MAGIC_ROOM:
		{
			ROOM_INDEX_DATA *room = ptr;
			snprintf(buf, 256, "[%5d] %s", room->vnum, nocolor(room->name));
			break;
		}
		
		case MAGIC_OBJ:
		{
			OBJ_DATA *obj = ptr;
			snprintf(buf, 256, "[%5d] %s", obj->pIndexData->vnum, nocolor(obj->short_descr));
			break;
		}
		
		case MAGIC_AREA:
		{
			AREA_DATA *area = ptr;
			snprintf(buf, 256, "[%5d] %s", area->vnum, area->filename);
			break;
		}
		
		default:
			snprintf(buf, 256, "Unknown (%d)", *(unsigned char*)ptr);
	}

	BUFDONE(buf);
	return buf;
}

/* Describe event shortly */
const char * print_event (EVENT *e)
{
    char *buf=BUF(256);
	int left = (e->when - sysdata.pulses) / PULSE_PER_SECOND;

	/* Hmm, this assumes sizeof (any*)	 == sizeof(void*) */
	sprintf(buf, "%-25.25s %3d:%02d %-5s ", e->actor_type == eva_global ? "Global" : ptr_string(e->actor.typeless), left / 60 , left % 60,
		flag_string(event_types, e->type)+6);

	if (e->type == event_misc)
		sprintf(buf+strlen(buf), "%s %d %d %d %d",
			flag_string(event_misc_types, e->subtype)+6,
			e->data.misc.num[0],
			e->data.misc.num[1],
			e->data.misc.num[2],
			e->data.misc.num[3]);
	else if (e->type == event_temp)
		sprintf(buf+strlen(buf), "%s",
			flag_string(event_temp_types, e->subtype)+6);
	else if (e->type == event_mprog)
		sprintf(buf+strlen(buf), "%s %d",
			e->data.mprog.name,
			e->data.mprog.vnum);

	else if (e->type == event_action)			
		sprintf(buf+strlen(buf), "%s", action_table[e->subtype].description);
	
	return buf;
}

/* Update/execute events */
void event_update()
{
	EVENT *e;
    ROOM_INDEX_DATA *r;
	
	DEBF(); /* this will added the name of the current function to a debug buffer (which gets written out to a file if the game crashes */

    /* As long as there are more vents in the current hash bucket that
     * are to go off this pulse
     */
	while ((e = event_hash[(sysdata.pulses) % MAX_EVENT_HASH]) && e->when == sysdata.pulses)
	{
		r=NULL;
		
		/* Unlink the event before doing anything */
		event_hash[(sysdata.pulses) % MAX_EVENT_HASH] = e->next_global;
		
        /* Remove the event from its object (list threaded on next_local)
         */
		switch (e->actor_type)
		{
			case eva_obj:
				remove_event_local(&e->actor.obj->event, e);
				break;

			case eva_ch:
				remove_event_local(&e->actor.ch->event, e);
				break;

			case eva_room:
                r=e->actor.room;
				remove_event_local(&e->actor.room->event, e);
				break;

			case eva_area:
				remove_event_local(&e->actor.area->event, e);
				break;
			
            case eva_global: /* No local queue for these */
				remove_event_local(&global_events, e);
				break;
		}

#ifdef EVENT_DEBUG
	fprintf(stdout, "-- %s\n", print_event(e));
#endif	
		
		/* Do stuff with e */
		switch (e->type)
		{
			default:
				do_abort(); /* unknown event type */
			
			case event_mprog: /* Call a mob prog */
				if (e->data.mprog.vnum) /* Call a mprog or just symbolic event? */
				{
					MPROG_CODE *code = get_mprog_index(e->data.mprog.vnum);

                    /* Higly specific mob prog stuff deleted
                     */
                    
                }
                break;

                /* Temporary events have a function pointer that should be called */
            case event_temp:
            if (e->data.temp.fun)
					e->data.temp.fun (e);
                break;

                /* Action events update ch->action: a character can perform one action at a time (e.g. spellcasting) */
            case event_action:
				e->actor.ch->action = NULL;
				action_table[e->subtype].fun(e);
				free_string(e->data.action.args);
				break;

                /* Misc events save with players/objects */
			case event_misc: /* Something else, usually nothing */
				if (misc_event_callbacks[e->subtype])
					misc_event_callbacks[e->subtype] (e);
				break;
		}

		free_event(e);
		sysdata.events.removed++;

#ifdef OH /* something to do with the virtualization of the wilderness, blame Oliver :) */
		if (r && !r->event)
			virtual_check(r);
#endif
	}
}

/* Add an event to the global list */
void event_add (EVENT *e)
{
	EVENT **ee;

    /* Soemone tried to add an event that was to go off BEFORE now? */
	if (e->when < sysdata.pulses)
	{
		fprintf(stderr, "** BD EVEN:%s\n", print_event(e));
		e->when = sysdata.pulses + 1;
	}

    /* I had a lot of trouble with doubly added events, so a sanity check here */
	if (e->added)
	{
		bugf ("Re-added already added event");
		do_abort();
	}
	
	e->added = TRUE;

    /* Where is the hash bucket we want to add this event to? */
	ee = &event_hash[(e->when) % MAX_EVENT_HASH];

    /* Nothing there, or the first event there goes off after ours */
	if (*ee == NULL || (*ee)->when > e->when)
	{
		e->next_global = *ee;
		*ee = e;
	}
	else /* go through that list, and find the event to go off before ours */
	{
		EVENT *prev;
		for (prev = *ee; prev->next_global && prev->next_global->when <= e->when; prev = prev->next_global)
			;

		e->next_global = prev->next_global;
		prev->next_global = e;
	}

#ifdef EVENT_DEBUG
	fprintf(stdout, "++ %s\n", print_event(e));
#endif	

    /* Some statistics */
	if (++sysdata.events.added-sysdata.events.removed > sysdata.events.highwater)
		sysdata.events.highwater = sysdata.events.added - sysdata.events.removed;
}

/* Find and remove this event from the global list */
void event_remove (EVENT *e)
{
    /* Head of the hash bucket? */
	if (event_hash[e->when % MAX_EVENT_HASH] == e)
		event_hash[e->when % MAX_EVENT_HASH] = e->next_global;
	else
	{
		EVENT *prev;
		
		for (prev = event_hash[e->when % MAX_EVENT_HASH]; prev; prev = prev->next_global)
			if (prev->next_global == e)
				break;
				
		if (prev)
			prev->next_global = e->next_global;
		else /* Ouch, this shouldn't happen! */
			fprintf(stderr, "?? %s\n", print_event(e));
	}

#ifdef EVENT_DEBUG
	fprintf(stdout, "// %s\n", print_event(e));
#endif	

    sysdata.events.removed++;
    e->added = FALSE;
}

void event_free (EVENT *e)
{
    /* If marked as loading, we have not injected it in the global list yet
     * Events on charaters that are not fully loaded just stay on their local
     * lists, until they enter the game, then ch_inject_events gets called
     */
	if (!e->loading)
		event_remove(e);
	
#ifdef EVENT_DEBUG
	fprintf(stdout, "== %s\n", print_event(e));
#endif

    free_event(e);
}

/* Look for a misc event
 * Note that this takes the head of the local event list, e.g. you'd do:
 * EVENT *e = get_misc_event(ch->event (head of the events on ch) , emisc_something, 0);
 */
EVENT *get_misc_event (EVENT *e, int subtype, int value0)
{
	for (; e; e = e->next_local)
		if (e->type == event_misc && e->subtype == subtype
		    && e->data.misc.num[0] == value0)
		return e;
	
	return NULL;
}

/* Does ch have a temp event with subtype subtype on him, and data == data? */
EVENT* get_temp_event(EVENT *e, int subtype, int key)
{
	for (; e; e = e->next_local)
		if (e->type == event_temp && e->subtype == subtype
		    && e->data.temp.key == key)
		return e;
	
	return NULL;
}

/* Find time left on this event */
int get_time_left (EVENT *e, int type, int subtype)
{
	for (; e ; e = e->next_local)
		if (e->type == type && e->subtype == subtype)
			return e->when - sysdata.pulses;
	
	return -1;
}


/* Save this event to a file */
void save_event (FILE *fp, const EVENT *e)
{
    /* Don't save those events */
	if (e->temporary || e->no_quit || e->type == event_temp || e->type == event_action)
		return;

    /* Eek! a buggy event somehow got here! */
	if (!e->loading && e->when < sysdata.pulses)
		return;
		
	fprintf (fp, "Event %d %d %d ",
		e->loading ? e->when : e->when - sysdata.pulses, e->type, e->subtype);
		
	switch (e->type)
	{
		default: /* Bad event, should never save! */
			do_abort();

		/* Although some event subtypes may define an overlaying structure,
		   this structure may only contain ints */			
		case event_misc:
			fprintf (fp, "%d %d %d %d\n",
				e->data.misc.num[0],
				e->data.misc.num[1],
				e->data.misc.num[2],
				e->data.misc.num[3]);
			break;

            /* mprog events have their own categoery because they save the mprog name */
		case event_mprog:
			fprintf(fp, "%s~ %d\n", e->data.mprog.name, e->data.mprog.vnum);
			break;
	}
}

/* Load an event. it's assumed that we are standing just right after "event" */
EVENT * load_event (FILE *fp)
{
	EVENT *e = new_event();
	int i;
	
	e->when = fread_number (fp);

		
	e->type = fread_number (fp);
	e->subtype = fread_number (fp);
	e->loading = TRUE;

	switch (e->type)
	{
		default:
			do_abort();
		
		case event_misc:
			for (i = 0; i < MAX_MISC_DATA; i++)
				e->data.misc.num[i] = fread_number(fp);
			break;
		
		case event_mprog:
			e->data.mprog.name = fread_string(fp);
			e->data.mprog.vnum = fread_number(fp);
			break;
	}

	if (e->when < 0)
	{
		bugf ("Bad event: %s", print_event(e));
		e->when = 42;
	}
	
	return e;
}

/* Set the timer of this object (pulses!) */
void set_timer (OBJ_DATA *obj, int pulses)
{	
	EVENT *e;

    /* negative time? */
	if (pulses < 1)
	{
		bugf ("set_timer called with pulses=%d; [%5d] %s", pulses, obj->pIndexData->vnum, obj->short_descr);
		pulses = 4;
	}

    /* deleted object? */
	if (obj->deleted)
	{
		FORK_ABORT();
		return;
	}

    /* Do we have the same event there already? */
	for (e = obj->event; e; e = e->next_local)
		if (e->type == event_misc && e->subtype == emisc_object_decay)
			break;

	if (e)
		event_remove (e); /* Get rid of the existing event */
	else
	{
		e = new_event();
		e->type = event_misc;
		e->actor_type = eva_obj;
		e->actor.obj = obj;
		e->subtype = emisc_object_decay;
		e->next_local = obj->event;
		obj->event = e;
	}

	e->when = sysdata.pulses + pulses;
	event_add(e);
}

/* Insert a very simple event */
EVENT * insert_simple_event (CHAR_DATA *ch, int type, int subtype, int seconds)
{
	EVENT *e = new_event();
	
	if (ch->deleted)
        do_abort();
	
	e->type = type;
	e->subtype = subtype;
	e->when = sysdata.pulses + seconds * PULSE_PER_SECOND;
	e->actor_type = eva_ch;
	e->actor.ch = ch;
	e->next_local = ch->event;
	ch->event = e;
	
    event_add(e);

    return e;
}

/* Insert a very simple event */
EVENT* insert_simple_obj_event (OBJ_DATA *obj, int type, int subtype, int seconds)
{
	EVENT *e;
	
    if (obj->deleted)
        do_abort();
	
	for (e = obj->event; e; e = e->next_local)
		if (e->type == type && e->subtype == subtype)
			break;
		
	if (e)
	{
		event_remove(e);
		remove_event_local(&obj->event, e);
	}
	else
		e = new_event();
		
	e->type = type;
	e->subtype = subtype;
	e->when = sysdata.pulses + seconds * PULSE_PER_SECOND;
	e->actor_type = eva_obj;
	e->actor.obj = obj;
	e->next_local = obj->event;
	obj->event = e;
	
	event_add(e);
	return e;
}

/* Easy setting of timer in ticks */
void set_timer_tick(OBJ_DATA *obj, int ticks)
{
	set_timer(obj, percent_fuzzy(ticks * PULSE_TICK,5));
}

void obj_add_triggers (OBJ_DATA *obj);


/* Event injection: moving of events on a loading character into the global
 * list when he enters the game
 */
void obj_inject_events (OBJ_DATA *obj)
{
	EVENT *e;
    OBJ_DATA *obj_inside;

    obj->key = generate_key(obj);

	for (e = obj->event; e; e = e->next_local)
	{
		e->when += sysdata.pulses; /* Events are saved as offsets */
		event_add(e);
		e->loading = 0;
		
	}

    for (obj_inside = obj->contains; obj_inside; obj_inside = obj_inside->next_content)
        obj_inject_events(obj_inside);

    obj_add_triggers (obj); /* Always add trigger AFTER loaded triggers are activated! */
}

/* Activate events on characters and his objects; move into global queue */
void inject_events (CHAR_DATA *ch)
{
	OBJ_DATA *obj_inside;
	CORPSE_DATA *corpse;
	EVENT *e;
	
	for (e = ch->event; e; e = e->next_local)
	{
		e->when += sysdata.pulses; /* Events are saved as offsets */
		event_add(e);
		e->loading = 0;
	}
	
	for (obj_inside = ch->carrying; obj_inside; obj_inside = obj_inside->next_content)
		obj_inject_events(obj_inside);

	for (corpse = ch->pcdata->corpses; corpse; corpse = corpse->next)
		obj_inject_events(corpse->obj);
}	

void do_dump (CHAR_DATA *ch, char *argument, void *target)
{
    BUFFER *buffer = buffer_new (10000);
    int i,last;
    EVENT *e;
    
    if (!str_cmp(argument, "list"))
    {
        for (i = 0; i < MAX_EVENT_HASH; i++)
        {
            last = 0;
            for (e = event_hash[i]; e; e = e->next_global)
            {
                bprintf(buffer, "%s\n", print_event(e));
                last = e->when;
            }
        }
    }
    else if (!str_cmp(argument, "stats"))
    {
        long total=0;
        long ms=0;
        long len;
        
        for (i=0; i<MAX_EVENT_HASH; i++)
        {
            len=0;
            for (e=event_hash[i]; e; e=e->next_global)
            {
                total++;
                len++;
            }
            
            ms += len*len;
        }
        
        bprintf(buffer, "Mean-squared hash bucket size: %f\n", (float)ms / (float)total / 2.0);
    }
    else
    {
        buffer_strcat(buffer, "Dump list (large) or stats?\n");
    }
    
    send_to_char (buffer->data,ch);
    buffer_free(buffer);
    
}

EVENT* add_generic_callback(void *actor, int actor_type, EVENT **event_list, int when, int type, EventCallbackFun *fun, void *data)
{
	EVENT *e;

	/* Replace existing events */
	
	for (e = *event_list; e; e = e->next_local)
		if (e->type == event_temp && e->subtype == type)
			break;

	if (!e || type == etemp_stackable)
		e = new_event();
	else
	{
		event_remove (e);
		remove_event_local(event_list, e);
	}

	e->when = sysdata.pulses + when;
	e->type = event_temp;
	e->subtype = type;
	e->actor.typeless = actor;
	e->actor_type = actor_type;
	e->data.temp.fun = fun;
	e->data.temp.data = data;
	e->data.temp.key = 0;
	e->next_local = *event_list;
	*event_list = e;
	event_add(e);
	
	return e;
}

/* Generic flag-removal function */
void flag_removal (EVENT *e)
{
	switch (e->actor_type)
	{
		case eva_room:
			REMOVE_BIT(e->actor.room->room_flags, e->data.temp.key);
			break;
		
		default:
			;
	}
}

/* Send a message */
void send_string (EVENT *e)
{
	switch(e->actor_type)
	{
		case eva_ch:
			send_to_char(e->data.temp.data, e->actor.ch);
			break;
		default:
			;
	}
	
	free_string(e->data.temp.data);
}

void send_act (EVENT *e)
{
	switch (e->actor_type)
	{
		case eva_ch:
			act (e->data.temp.data, e->actor.ch, NULL, NULL, TO_ALL);
			break;
		
		case eva_room:	
			if (e->actor.room->people)
				act (e->data.temp.data, e->actor.room->people, NULL, NULL, TO_ALL);
			break;

		default:
			;
	}
}

EVENT* add_room_callback(ROOM_INDEX_DATA *room, int when, int type, EventCallbackFun *fun, void *data)
{	
	return add_generic_callback(room,eva_room, &room->event, when,type, fun,data);
}

EVENT* add_obj_callback(OBJ_DATA *obj, int when, int type, EventCallbackFun *fun, void *data)
{	
	if (obj->deleted)
	{
		FORK_ABORT();
		return NULL;
	}
	return add_generic_callback(obj,eva_obj, &obj->event, when,type, fun,data);
}

EVENT* add_ch_callback(CHAR_DATA *ch, int when, int type, EventCallbackFun *fun, void *data)
{	
	if (ch->deleted)
	{
		FORK_ABORT();
		return NULL;
	}
	return add_generic_callback(ch,eva_ch, &ch->event, when,type, fun,data);
}

EVENT* add_area_callback(AREA_DATA *area, int when, int type, EventCallbackFun *fun, void *data)
{	
	return add_generic_callback(area,eva_area, &area->event, when,type, fun,data);
}

EVENT* add_global_callback(int when, int type, EventCallbackFun *fun, void *data)
{	
	return add_generic_callback(global_events,eva_global, &global_events, when,type, fun,data);
}

/* Remove the event of this type from the list, if any */
void strip_event (EVENT **e_list, int type, int subtype)
{
	EVENT *e;
	
	for (e = *e_list; e; e = e->next_local)
		if (e->type == type && e->subtype == subtype)
		{
			event_remove(e);
            remove_event_local(e_list,e);
            free_event(e);
			break;
		}
}

/* Add an action */
EVENT* add_action (CHAR_DATA *ch, int sn, int when, int target, char *args)
{
	EVENT *e = new_event();
	
	if (ch->deleted)
	{
		FORK_ABORT();
		return NULL;
	}

	e->type = event_action;
	e->subtype = sn;
	e->when = sysdata.pulses + when;
	e->data.action.args = str_dup(args);
	e->data.action.target = target;
	e->data.action.duration = when;
	e->actor_type = eva_ch;
	e->actor.ch = ch;
	
	e->next_local = ch->event;
	ch->event = e;
	ch->action = e;
	
	event_add(e);
	
	return e;
}








